#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Caner Candan <caner@candan.fr>, http://caner.candan.fr
#

from duniterpy.api.bma import API, logging, parse_response

logger = logging.getLogger("duniter/node")

URL_PATH = 'node'

async def summary(connection):
    """
    GET Certification data over a member

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """
    schema = {
        "type": "object",
        "properties": {
            "duniter": {
                "type": "object",
                "properties": {
                    "software": {
                    "type": "string"
                    },
                    "version": {
                        "type": "string",
                    },
                    "forkWindowSize": {
                        "type": "number"
                    }
                },
                "required": ["software", "version"]
            },
        },
        "required": ["duniter"]
    }
    client = API(connection, URL_PATH)

    r = await client.requests_get('/summary')
    return await parse_response(r, schema)




class DuniterError(Exception):
    """
    duniter error
    """
    def __init__(self, data):
        super().__init__("Error code {0} - {1}".format(data["ucode"], data["message"]))
        self.ucode = data["ucode"]
        self.message = data["message"]

UNKNOWN                              = 1001
UNHANDLED                            = 1002
SIGNATURE_DOES_NOT_MATCH             = 1003
ALREADY_UP_TO_DATE                   = 1004
WRONG_DOCUMENT                       = 1005
HTTP_LIMITATION                      = 1006

HTTP_PARAM_PUBKEY_REQUIRED           = 1101
HTTP_PARAM_IDENTITY_REQUIRED         = 1102
HTTP_PARAM_PEER_REQUIRED             = 1103
HTTP_PARAM_BLOCK_REQUIRED            = 1104
HTTP_PARAM_MEMBERSHIP_REQUIRED       = 1105
HTTP_PARAM_TX_REQUIRED               = 1106
HTTP_PARAM_SIG_REQUIRED              = 1107
HTTP_PARAM_CERT_REQUIRED             = 1108
HTTP_PARAM_REVOCATION_REQUIRED       = 1109
HTTP_PARAM_CONF_REQUIRED             = 1110

NO_MATCHING_IDENTITY                 = 2001
UID_ALREADY_USED                     = 2002
PUBKEY_ALREADY_USED                  = 2003
NO_MEMBER_MATCHING_PUB_OR_UID        = 2004
SELF_PEER_NOT_FOUND                  = 2005
WRONG_SIGNATURE_MEMBERSHIP           = 2006
ALREADY_RECEIVED_MEMBERSHIP          = 2007
MEMBERSHIP_A_NON_MEMBER_CANNOT_LEAVE = 2008
NOT_A_MEMBER                         = 2009
NO_CURRENT_BLOCK                     = 2010
BLOCK_NOT_FOUND                      = 2011
PEER_NOT_FOUND                       = 2012
WRONG_UNLOCKER                       = 2013
LOCKTIME_PREVENT                     = 2014
SOURCE_ALREADY_CONSUMED              = 2015
WRONG_AMOUNTS                        = 2016
WRONG_OUTPUT_BASE                    = 2017
CANNOT_ROOT_BLOCK_NO_MEMBERS         = 2018


"""
duniter public and private keys

@author: inso
"""

import base58
import base64
import libnacl.sign
from pylibscrypt import scrypt
from .base58 import Base58Encoder


class VerifyingKey(libnacl.sign.Verifier):
    """
    Class to verify documents
    """
    def __init__(self, pubkey):
        """
        Creates a Verify class from base58 pubkey
        :param pubkey:
        """
        key = libnacl.encode.hex_encode(Base58Encoder.decode(pubkey))
        super().__init__(key)

    def verify_document(self, document, **kwargs):
        """
        Check specified document
        :param duniterpy.documents.Document document:
        :return:
        """
        signature = base64.b64decode(document.signatures[0])
        prepended = signature + bytes(document.raw(**kwargs), 'ascii')

        try:
            self.verify(prepended)
            return True
        except ValueError:
            return False



#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Caner Candan <caner@candan.fr>, http://caner.candan.fr
#

from duniterpy.api.bma import API, logging, parse_response

logger = logging.getLogger("duniter/tx")

URL_PATH = 'tx'

HISTORY_SCHEMA = {
    "type": "object",
    "properties": {
        "currency": {
            "type": "string"
        },
        "pubkey": {
            "type": "string"
        },
        "history": {
            "type": "object",
            "properties": {
                "sent": {
                    "$ref": "#/definitions/transaction_data"
                },
                "received": {
                    "$ref": "#/definitions/transaction_data"
                },
                "sending": {
                    "$ref": "#/definitions/transactioning_data"
                },
                "receiving": {
                    "$ref": "#/definitions/transactioning_data"
                },
            },
            "required": ["sent", "received", "sending", "receiving"]
        }
    },
    "definitions": {
        "transaction_data": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "version": {
                        "type": "number"
                    },
                    "issuers": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "inputs": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "outputs": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "unlocks": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "comment": {
                        "type": "string"
                    },
                    "signatures": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "hash": {
                        "type": "string"
                    },
                    "block_number": {
                        "type": "number"
                    },
                    "time": {
                        "type": "number"
                    }
                },
                "required": ["version", "issuers", "inputs", "outputs",
                             "comment", "signatures", "hash", "block_number", "time"]
            }
        },
        "transactioning_data": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "version": {
                        "type": "number"
                    },
                    "issuers": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "inputs": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "outputs": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "unlocks": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "comment": {
                        "type": "string"
                    },
                    "signatures": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "hash": {
                        "type": "string"
                    },
                },
                "required": ["version", "issuers", "inputs", "outputs",
                             "comment", "signatures", "hash"]
            }
        }
    },
    "required": ["currency", "pubkey", "history"]
}

SOURCES_SCHEMA = {
        "type": "object",
        "properties": {
            "currency": {
                "type": "string"
            },
            "pubkey": {
                "type": "string"
            },
            "sources": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "type": {
                            "type": "string"
                        },
                        "noffset": {
                            "type": "number"
                        },
                        "identifier": {
                            "type": "string"
                        },
                        "amount": {
                            "type": "number"
                        },
                        "base": {
                            "type": "number"
                        }
                    },
                    "required": ["type", "noffset", "identifier", "amount", "base"]
                }
            }
        },
        "required": ["currency", "pubkey", "sources"]
    }

async def history(connection, pubkey):
    """
    Get transactions history of public key

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str pubkey: Public key
    :rtype: dict
    """

    client = API(connection, URL_PATH)

    r = await client.requests_get( '/history/%s' % pubkey)
    return await parse_response(r, HISTORY_SCHEMA)

async def process(connection, transaction):
    """
    POST a transaction

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param duniterpy.documents.Transaction transaction: Transaction document
    :rtype: aiohttp.ClientResponse
    """
    client = API(connection, URL_PATH)

    r = await client.requests_post('/process', transaction=transaction)
    return r


async def sources(connection, pubkey):
    """
    GET transaction sources

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str pubkey: Public key
    :rtype: dict
    """
    client = API(connection, URL_PATH)

    r = await client.requests_get( '/sources/%s' % pubkey)
    return await parse_response(r, SOURCES_SCHEMA)

async def blocks(connection, pubkey, start, end):
    """
    GET public key transactions history between start and end block number

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str pubkey: Public key
    :param int start: Start from block number
    :param int end: End to block number
    :return: dict
    """
    client = API(connection, URL_PATH)

    r = await client.requests_get('/history/%s/blocks/%s/%s' % (pubkey, start, end))
    return await parse_response(r, HISTORY_SCHEMA)

async def times(connection, pubkey, start, end):
    """
    GET public key transactions history between start and end timestamp

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str pubkey: Public key
    :param int start: Start from timestamp
    :param int end: End to timestamp
    :return: dict
    """
    client = API(connection, URL_PATH)

    r = await client.requests_get('/history/%s/times/%s/%s' % (pubkey, start, end))
    return await parse_response(r, HISTORY_SCHEMA)



import aiohttp
import unittest
import jsonschema
import json
from tests.api.webserver import WebFunctionalSetupMixin, web, asyncio
from duniterpy.documents import BMAEndpoint
from duniterpy.api.bma.wot import lookup, members, certified_by, certifiers_of, requirements, \
    REQUIREMENTS_SCHEMA, CERTIFICATIONS_SCHEMA, LOOKUP_SCHEMA, MEMBERS_SCHEMA


class Test_BMA_Wot(WebFunctionalSetupMixin, unittest.TestCase):
    def test_bma_wot_lookup(self):
        json_sample = {
            "partial": False,
            "results": [
                {
                    "pubkey": "HsLShAtzXTVxeUtQd7yi5Z5Zh4zNvbu8sTEZ53nfKcqY",
                    "uids": [
                        {
                            "uid": "udid2;c;TOCQUEVILLE;FRANCOIS-XAVIER-ROBE;1989-07-14;e+48.84+002.30;0;",
                            "meta": {
                                "timestamp": "44-76522E321B3380B058DB6D9E66121705EEA63610869A7C5B3E701CF6AF2D55A8"
                            },
                            "self": "J3G9oM5AKYZNLAB5Wx499w61NuUoS57JVccTShUbGpCMjCqj9yXXqNq7dyZpDWA6BxipsiaMZhujMeBfCznzyci",
                            "others": [
                                {
                                    "pubkey": "9WYHTavL1pmhunFCzUwiiq4pXwvgGG5ysjZnjz9H8yB",
                                    "meta": {
                                        "timestamp": "44-76522E321B3380B058DB6D9E66121705EEA63610869A7C5B3E701CF6AF2D55A8"
                                    },
                                    "signature": "42yQm4hGTJYWkPg39hQAUgP6S6EQ4vTfXdJuxKEHL1ih6YHiDL2hcwrFgBHjXLRgxRhj2VNVqqc6b4JayKqTE14r"
                                }
                            ]
                        }
                    ],
                    "signed": [
                        {
                            "uid": "snow",
                            "pubkey": "2P7y2UDiCcvsgSSt8sgHF3BPKS4m9waqKw4yXHCuP6CN",
                            "meta": {
                                "timestamp": "44-76522E321B3380B058DB6D9E66121705EEA63610869A7C5B3E701CF6AF2D55A8"
                            },
                            "signature": "Xbr7qhyGNCmLoVuuKnKIbrdmtCvb9VBIEY19izUNwA5nufsjNm8iEsBTwKWOo0lq5O1+AAPMnht8cm2JjMq8AQ=="
                        },
                        {
                            "uid": "snow",
                            "pubkey": "2P7y2UDiCcvsgSSt8sgHF3BPKS4m9waqKw4yXHCuP6CN",
                            "meta": {
                                "timestamp": "44-76522E321B3380B058DB6D9E66121705EEA63610869A7C5B3E701CF6AF2D55A8"
                            },
                            "signature": "HU9VPwC4EqPJwATPuyUJM7HLjfig5Ke1CKonL9Q78n5/uNSL2hkgE9Pxsor8CCJfkwCxh66NjGyqnGYqZnQMAg=="
                        },
                        {
                            "uid": "snow",
                            "pubkey": "7xapQvvxQ6367bs8DsskEf3nvQAgJv97Yu11aPbkCLQj",
                            "meta": {
                                "timestamp": "44-76522E321B3380B058DB6D9E66121705EEA63610869A7C5B3E701CF6AF2D55A8"
                            },
                            "signature": "6S3x3NwiHB2QqYEY79x4wCUYHcDctbazfxIyxejs38V1uRAl4DuC8R3HJUfD6wMSiWKPqbO+td+8ZMuIn0L8AA=="
                        },
                        {
                            "uid": "cat",
                            "pubkey": "CK2hBp25sYQhdBf9oFMGHyokkmYFfzSCmwio27jYWAN7",
                            "meta": {
                                "timestamp": "44-76522E321B3380B058DB6D9E66121705EEA63610869A7C5B3E701CF6AF2D55A8"
                            },
                            "signature": "AhgblSOdxUkLwpUN9Ec46St3JGaw2jPyDn/mLcR4j3EjKxUOwHBYqqkxcQdRz/6K4Qo/xMa941MgUp6NjNbKBA=="
                        }
                    ]
                }
            ]
        }
        jsonschema.validate(json_sample, LOOKUP_SCHEMA)

    def test_bma_wot_lookup_bad(self):
        async def handler(request):
            await request.read()
            return web.Response(body=b'{}', content_type='application/json')

        async def go():
            _, srv, port, url = await self.create_server('GET', '/wot/lookup/pubkey', handler)
            with self.assertRaises(jsonschema.exceptions.ValidationError):
                with aiohttp.ClientSession() as session:
                    connection = next(BMAEndpoint("127.0.0.1", None, None, port).conn_handler(session))
                    await lookup(connection, 'pubkey')

        self.loop.run_until_complete(go())

    def test_bma_wot_members(self):
        json_sample = {
            "results": [
                {"pubkey": "HsLShAtzXTVxeUtQd7yi5Z5Zh4zNvbu8sTEZ53nfKcqY", "uid": "cat"},
                {"pubkey": "9kNEiyseUNoPn3pmNUhWpvCCwPRgavsLu7YFKZuzzd1L", "uid": "tac"},
                {"pubkey": "9HJ9VXa9wc6EKC6NkCi8b5TKWBot68VhYDg7kDk5T8Cz", "uid": "toc"}
            ]
        }
        jsonschema.validate(MEMBERS_SCHEMA, json_sample)

    def test_bma_wot_members_bad(self):
        async def handler(request):
            await request.read()
            return web.Response(body=b'{}', content_type='application/json')

        async def go():
            _, srv, port, url = await self.create_server('GET', '/wot/members', handler)
            with self.assertRaises(jsonschema.exceptions.ValidationError):
                with aiohttp.ClientSession() as session:
                    connection = next(BMAEndpoint("127.0.0.1", None, None, port).conn_handler(session))
                    await members(connection)

        self.loop.run_until_complete(go())

    def test_bma_wot_cert(self):
        json_sample = {
            "pubkey": "HsLShAtzXTVxeUtQd7yi5Z5Zh4zNvbu8sTEZ53nfKcqY",
            "uid": "user identifier",
            "isMember": True,
            "certifications": [
                {
                    "pubkey": "9WYHTavL1pmhunFCzUwiiq4pXwvgGG5ysjZnjz9H8yB",
                    "uid": "certifier uid",
                    "cert_time": {
                        "block": 88,
                        "medianTime": 1509991044
                    },
                    "sigDate": "80-D30978C9D6C5A348A8188603F039423D90E50DC5",
                    "written": {
                        "number": 872768,
                        "hash": "D30978C9D6C5A348A8188603F039423D90E50DC5"
                    },
                    "isMember": True,
                    "wasMember": True,
                    "signature": "42yQm4hGTJYWkPg39hQAUgP6S6EQ4vTfXdJuxKEHL1ih6YHiDL2hcwrFgBHjXLRgxRhj2VNVqqc6b4JayKqTE14r"
                },
                {
                    "pubkey": "9WYHTavL1pmhunFCzUwiiq4pXwvgGG5ysjZnjz9H8yB",
                    "uid": "certifier uid",
                    "sigDate": "80-D30978C9D6C5A348A8188603F039423D90E50DC5",
                    "cert_time": {
                        "block": 88,
                        "medianTime": 1509991044
                    },
                    "written": None,
                    "isMember": True,
                    "wasMember": False,
                    "signature": "42yQm4hGTJYWkPg39hQAUgP6S6EQ4vTfXdJuxKEHL1ih6YHiDL2hcwrFgBHjXLRgxRhj2VNVqqc6b4JayKqTE14r"
                }
            ]
        }
        jsonschema.validate(json_sample, CERTIFICATIONS_SCHEMA)
        jsonschema.validate(json_sample, CERTIFICATIONS_SCHEMA)

    def test_bma_wot_certifiers_bad(self):
        async def handler(request):
            await request.read()
            return web.Response(body=b'{}', content_type='application/json')

        async def go():
            _, srv, port, url = await self.create_server('GET', '/wot/certifiers-of/pubkey', handler)
            with self.assertRaises(jsonschema.exceptions.ValidationError):
                with aiohttp.ClientSession() as session:
                    connection = next(BMAEndpoint("127.0.0.1", None, None, port).conn_handler(session))
                    await certifiers_of(connection, 'pubkey')

        self.loop.run_until_complete(go())

    def test_bma_wot_certifiers_inner_bad(self):
        async def handler(request):
            await request.read()
            return web.Response(body=bytes(json.dumps({
                "pubkey": "7Aqw6Efa9EzE7gtsc8SveLLrM7gm6NEGoywSv4FJx6pZ",
                "uid": "john",
                "isMember": True,
                "certifications": [
                    {
                        "pubkey": "FADxcH5LmXGmGFgdixSes6nWnC4Vb4pRUBYT81zQRhjn",
                        "meta": {
                            "block_number": 38580
                        },
                        "uids": [
                            "doe"
                        ],
                        "isMember": True,
                        "wasMember": True,
                        "signature": "8XYmBdElqNkkl4AeFjJnC5oj/ujBrzH9FNgPZvK8Cicp8Du0PQa0yYFG95EQ46MJhdV0fUT2g5xyH8N3/OGhDA=="
                    },
                ]
            }), "utf-8"), content_type='application/json')

        async def go():
            _, srv, port, url = await self.create_server('GET', '/wot/certifiers-of/pubkey', handler)
            with self.assertRaises(jsonschema.exceptions.ValidationError):
                with aiohttp.ClientSession() as session:
                    connection = next(BMAEndpoint("127.0.0.1", None, None, port).conn_handler(session))
                    await certifiers_of(connection, 'pubkey')

        self.loop.run_until_complete(go())

    def test_bma_wot_certified_bad(self):
        async def handler(request):
            await request.read()
            return web.Response(body=b'{}', content_type='application/json')

        async def go():
            _, srv, port, url = await self.create_server('GET', '/wot/certified-by/pubkey', handler)
            with self.assertRaises(jsonschema.exceptions.ValidationError):
                with aiohttp.ClientSession() as session:
                    connection = next(BMAEndpoint("127.0.0.1", None, None, port).conn_handler(session))
                    await certified_by(connection, 'pubkey')

        self.loop.run_until_complete(go())

    def test_bma_wot_requirements(self):
        json_sample = {
            "identities": [
                {
                    "pubkey": "8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU",
                    "uid": "inso",
                    "meta": {
                        "timestamp": "1470-46221DE81776D8382F6DE595105386ADEDD291BEC33D238C506F56EA3721B012"
                    },
                    "outdistanced": False,
                    "revocation_sig": None,
                    "revoked": False,
                    "revoked_on": None,
                    "expired": False,
                    "isSentry": True,
                    "wasMember": True,
                    "certifications": [
                        {
                            "from": "J78bPUvLjxmjaEkdjxWLeENQtcfXm7iobqB49uT1Bgp3",
                            "to": "8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU",
                            "expiresIn": 30423649
                        },
                        {
                            "from": "9bZEATXBGPUSsk8oAYi4KAChg3rHKwNt67hVdErbNGCW",
                            "to": "8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU",
                            "expiresIn": 30488510
                        },
                        {
                            "from": "HGYV5C16mrdvE9vpb1S9nMDHkVPsubBgANs9pSb6HWCV",
                            "to": "8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU",
                            "expiresIn": 30505972
                        },
                        {
                            "from": "5ocqzyDMMWf1V8bsoNhWb1iNwax1e9M7VTUN6navs8of",
                            "to": "8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU",
                            "expiresIn": 30923721
                        }
                    ],
                    "membershipPendingExpiresIn": 0,
                    "membershipExpiresIn": 14707940
                }
            ]
        }
        jsonschema.validate(json_sample, REQUIREMENTS_SCHEMA)




import re

from ..api.bma import ConnectionHandler
from .document import Document, MalformedDocumentError
from . import BlockUID
from .. import MANAGED_API
from .constants import block_hash_regex, pubkey_regex, ipv4_regex, ipv6_regex


class Peer(Document):
    """
.. note:: A peer document is specified by the following format :

    | Version: VERSION
    | Type: Peer
    | Currency: CURRENCY_NAME
    | PublicKey: NODE_PUBLICKEY
    | Block: BLOCK
    | Endpoints:
    | END_POINT_1
    | END_POINT_2
    | END_POINT_3
    | [...]

    """

    re_type = re.compile("Type: (Peer)")
    re_pubkey = re.compile("PublicKey: ({pubkey_regex})\n".format(pubkey_regex=pubkey_regex))
    re_block = re.compile("Block: ([0-9]+-{block_hash_regex})\n".format(block_hash_regex=block_hash_regex))
    re_endpoints = re.compile("(Endpoints:)\n")

    fields_parsers = {**Document.fields_parsers, **{
        "Type": re_type,
        "Pubkey": re_pubkey,
        "Block": re_block,
        "Endpoints": re_endpoints
    }}

    def __init__(self, version, currency, pubkey, blockUID,
                 endpoints, signature):
        super().__init__(version, currency, [signature])

        self.pubkey = pubkey
        self.blockUID = blockUID
        self.endpoints = endpoints

    @classmethod
    def from_signed_raw(cls, raw):
        lines = raw.splitlines(True)
        n = 0

        version = int(Peer.parse_field("Version", lines[n]))
        n += 1

        Peer.parse_field("Type", lines[n])
        n += 1

        currency = Peer.parse_field("Currency", lines[n])
        n += 1

        pubkey = Peer.parse_field("Pubkey", lines[n])
        n += 1

        blockUID = BlockUID.from_str(Peer.parse_field("Block", lines[n]))
        n += 1

        Peer.parse_field("Endpoints", lines[n])
        n += 1

        endpoints = []
        while not Peer.re_signature.match(lines[n]):
            endpoints.append(endpoint(lines[n]))
            n += 1

        signature = Peer.re_signature.match(lines[n]).group(1)

        return cls(version, currency, pubkey, blockUID, endpoints, signature)

    def raw(self):
        doc = """Version: {0}
Type: Peer
Currency: {1}
PublicKey: {2}
Block: {3}
Endpoints:
""".format(self.version, self.currency, self.pubkey, self.blockUID)

        for endpoint in self.endpoints:
            doc += "{0}\n".format(endpoint.inline())

        return doc


def endpoint(value):
    if isinstance(value, UnknownEndpoint):
        return value
    elif isinstance(value, BMAEndpoint):
        return value
    elif isinstance(value, str):
        for api in MANAGED_API:
            if value.startswith(api):
                if api == "BASIC_MERKLED_API":
                    return BMAEndpoint.from_inline(value)
                if api == "BMAS":
                    return SecuredBMAEndpoint.from_inline(value)
        return UnknownEndpoint.from_inline(value)
    else:
        raise TypeError("Cannot convert {0} to endpoint".format(value))


class Endpoint():
    @classmethod
    def from_inline(cls, inline):
        raise NotImplementedError("from_inline(..) is not implemented")

    def inline(self):
        raise NotImplementedError("inline() is not implemented")

    def __str__(self):
        raise NotImplementedError("__str__ is not implemented")

    def __eq__(self, other):
        raise NotImplementedError("__eq__ is not implemented")


class UnknownEndpoint(Endpoint):
    def __init__(self, api, properties):
        self.api = api
        self.properties = properties

    @classmethod
    def from_inline(cls, inline):
        api = inline.split()[0]
        properties = inline.split()[1:]
        return cls(api, properties)

    def inline(self):
        doc = self.api
        for p in self.properties:
            doc += " {0}".format(p)
        return doc

    def __str__(self):
        return "{0} {1}".format(self.api, ' '.join(["{0}".format(p) for p in self.properties]))

    def __eq__(self, other):
        if isinstance(other, UnknownEndpoint):
            return self.api == other.api and self.properties == other.properties
        else:
            return False

    def __hash__(self):
        return hash((self.api, self.properties))


class BMAEndpoint(Endpoint):
    re_inline = re.compile('^BASIC_MERKLED_API(?: ([a-z0-9-_.]*(?:.[a-zA-Z])))?(?: ({ipv4_regex}))?(?: ({ipv6_regex}))?(?: ([0-9]+))$'.format(ipv4_regex=ipv4_regex,
                                                                                                                                            ipv6_regex=ipv6_regex))

    def __init__(self, server, ipv4, ipv6, port):
        self.server = server
        self.ipv4 = ipv4
        self.ipv6 = ipv6
        self.port = port

    @classmethod
    def from_inline(cls, inline):
        m = BMAEndpoint.re_inline.match(inline)
        if m is None:
            raise MalformedDocumentError("BMAEndpoint")
        server = m.group(1)
        ipv4 = m.group(2)
        ipv6 = m.group(3)
        port = int(m.group(4))
        return cls(server, ipv4, ipv6, port)

    def inline(self):
        return "BASIC_MERKLED_API{DNS}{IPv4}{IPv6}{PORT}" \
                    .format(DNS=(" {0}".format(self.server) if self.server else ""),
                            IPv4=(" {0}".format(self.ipv4) if self.ipv4 else ""),
                            IPv6=(" {0}".format(self.ipv6) if self.ipv6 else ""),
                            PORT=(" {0}".format(self.port) if self.port else ""))

    def conn_handler(self, session=None, proxy=None):
        """
        Return connection handler instance for the endpoint

        :param aiohttp.ClientSession session: AIOHTTP client session instance
        """
        if self.server:
            yield ConnectionHandler("http", "ws", self.server, self.port, "", proxy, session)
        elif self.ipv6:
            yield ConnectionHandler("http", "ws", "[{0}]".format(self.ipv6), self.port, "", proxy, session)
        else:
            yield ConnectionHandler("http", "ws", self.ipv4, self.port, "", proxy, session)

    def __str__(self):
        return self.inline()

    def __eq__(self, other):
        if isinstance(other, BMAEndpoint):
            return self.server == other.server and self.ipv4 == other.ipv4 \
                    and self.ipv6 == other.ipv6 and self.port == other.port
        else:
            return False

    def __hash__(self):
        return hash((self.server, self.ipv4, self.ipv6, self.port))


class SecuredBMAEndpoint(BMAEndpoint):
    re_inline = re.compile('^BMAS(?: ([a-z0-9-_.]*(?:.[a-zA-Z])))?(?: ({ipv4_regex}))?(?: ({ipv6_regex}))? ([0-9]+)(?: ([/\w \.-]*)/?)?$'.format(ipv4_regex=ipv4_regex,
                                                                                                                                                 ipv6_regex=ipv6_regex))

    def __init__(self, server, ipv4, ipv6, port, path):
        super().__init__(server, ipv4, ipv6, port)
        self.path = path

    @classmethod
    def from_inline(cls, inline):
        m = SecuredBMAEndpoint.re_inline.match(inline)
        if m is None:
            raise MalformedDocumentError("BMAS")
        server = m.group(1)
        ipv4 = m.group(2)
        ipv6 = m.group(3)
        port = int(m.group(4))
        path = m.group(5)
        if not path:
            path = ""
        return cls(server, ipv4, ipv6, port, path)

    def inline(self):
        inlined = [str(info) for info in (self.server, self.ipv4, self.ipv6, self.port, self.path) if info]
        return "BMAS " + " ".join(inlined)

    def conn_handler(self, session=None, proxy=None):
        """
        Return connection handler instance for the endpoint

        :param aiohttp.ClientSession session: AIOHTTP client session instance
        :rtype: ConnectionHandler
        """
        if self.server:
            yield ConnectionHandler("https", "wss", self.server, self.port, self.path, proxy, session)
        elif self.ipv6:
            yield ConnectionHandler("https", "wss", "[{0}]".format(self.ipv6), self.port, self.path, proxy, session)
        else:
            yield ConnectionHandler("https", "wss", self.ipv4, self.port, self.path, proxy, session)



#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Caner Candan <caner@candan.fr>, http://caner.candan.fr
#
from duniterpy.api.bma import API, logging, parse_response

logger = logging.getLogger("duniter/wot")

URL_PATH = 'wot'


CERTIFICATIONS_SCHEMA = {
    "type": "object",
    "properties": {
        "pubkey": {
            "type": "string"
        },
        "uid": {
            "type": "string"
        },
        "isMember": {
            "type": "boolean"
        },
        "certifications": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "pubkey": {
                        "type": "string"
                    },
                    "uid": {
                        "type": "string"
                    },
                    "cert_time": {
                        "type": "object",
                        "properties": {
                            "block": {
                                "type": "number"
                            },
                            "medianTime": {
                                "type": "number"
                            }
                        },
                        "required": ["block", "medianTime"]
                    },
                    "sigDate": {
                        "type": "string"
                    },
                    "written": {
                        "oneOf": [
                            {
                                "type": "object",
                                "properties": {
                                    "number": {
                                        "type": "number",
                                    },
                                    "hash": {
                                        "type": "string"
                                    }
                                },
                                "required": ["number", "hash"]
                            },
                            {
                                "type": "null"
                            }
                        ]
                    },
                    "isMember": {
                        "type": "boolean"
                    },
                    "wasMember": {
                        "type": "boolean"
                    },
                    "signature": {
                        "type": "string"
                    }
                },
                "required": ["pubkey", "uid", "cert_time", "sigDate",
                             "written", "wasMember", "isMember", "signature"]
            }
        }
    },
    "required": ["pubkey", "uid", "isMember", "certifications"]
}


MEMBERS_SCHEMA = {
    "type": "object",
    "properties": {
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "pubkey": {
                        "type": "string"
                    }
                },
                "required": ["pubkey"]
            }
        }
    },
    "required": ["results"]
}


REQUIREMENTS_SCHEMA = {
    "type": "object",
    "properties": {
        "identities": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "pubkey": {
                        "type": "string"
                    },
                    "uid": {
                        "type": "string"
                    },
                    "meta": {
                        "type": "object",
                        "properties": {
                            "timestamp": {
                                "type": "string"
                            }
                        },
                        "required": ["timestamp"]
                    },
                    "outdistanced": {
                        "type": "boolean"
                    },
                    "certifications": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "from": {
                                    "type": "string"
                                },
                                "to": {
                                    "type": "string"
                                },
                                "expiresIn": {
                                    "type": "number"
                                }
                            },
                            "required": ["from", "to", "expiresIn"]
                        }
                    },
                    "membershipPendingExpiresIn": {
                        "type": "number"
                    },
                    "membershipExpiresIn": {
                        "type": "number"
                    },
                    "wasMember": {
                        "type": "boolean"
                    },
                    "isSentry": {
                        "type": "boolean"
                    },
                    "revoked": {
                        "type": "boolean"
                    },
                    "revokation_sig": {
                        "type": ["string", "null"]
                    },
                    "revoked_on": {
                        "type": ["number", "null"]
                    },
                },
                "required": ["pubkey", "uid", "meta", "outdistanced", "certifications", "membershipPendingExpiresIn",
                             "membershipExpiresIn", "wasMember", "isSentry", "revoked", "revoked_on", "revocation_sig"]
            }
        }
    },
    "required": ["identities"]
}

LOOKUP_SCHEMA = {
    "type": "object",
    "definitions": {
        "meta_data": {
            "type": "object",
            "properties": {
                "timestamp": {
                    "type": "string"
                }
            }
        },
    },
    "properties": {
        "partial": {
            "type": "boolean"
        },
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "pubkey": {
                        "type": "string"
                    }
                },
                "uids": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "uid": {
                                "type": "string"
                            },
                            "meta": {
                                "$ref": "#/definitions/meta_data"
                            },
                            "self": {
                                "type": "string",
                            },
                            "revokation_sig": {
                                "type": ["string", "null"]
                            },
                            "revoked_on": {
                                "type": ["number", "null"]
                            },
                            "revoked": {
                                "type": "boolean"
                            },
                            "others": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "pubkey": {
                                            "type": "string",
                                        },
                                        "meta": {
                                            "$ref": "#/definitions/meta_data"
                                        },
                                        "signature": {
                                            "type": "string"
                                        }
                                    }
                                }
                            }
                        },
                        "required": ["uid", "meta", "self", "revokation_sig", "revoked", "others"]
                    }
                },
                "signed": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "uid": {
                                "type": "string"
                            },
                            "pubkey": {
                                "type": "string"
                            },
                            "meta": {
                                "$ref": "#/definitions/meta_data"
                            },
                            "signature": {
                                "type": "string"
                            },
                            "revokation_sig": {
                                "type": ["string", "null"]
                            },
                            "revoked_on": {
                                "type": ["number", "null"]
                            },
                            "revoked": {
                                "type": "boolean"
                            }
                        },
                        "required": ["uid", "pubkey", "meta", "revokation_sig", "revoked", "signature"]
                    }
                },
                "required": ["uids", "signed"]
            }
        }
    },
    "required": ["partial", "results"]
}

async def add(connection, identity):
    """
    POST identity document

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param duniterpy.documents.certification.Identity identity: Identity document
    :rtype: aiohttp.ClientResponse
    """
    client = API(connection, URL_PATH)

    r = await client.requests_post('/add', identity=identity)
    return r

async def certify(connection, cert):
    """
    POST certification document

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param duniterpy.documents.certification.Certification cert: Certification document
    :rtype: aiohttp.ClientResponse
    """
    client = API(connection, URL_PATH)

    r = await client.requests_post('/certify', cert=cert)
    return r

async def revoke(connection, revocation):
    """
    POST revocation document

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param duniterpy.documents.certification.Revocation revocation: Certification document
    :rtype: aiohttp.ClientResponse
    """
    client = API(connection, URL_PATH)

    r = await client.requests_post('/revoke', revocation=revocation)
    return r


async def lookup(connection, search):
    """
    GET UID/Public key data

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str search: UID or public key
    :rtype: dict
    """
    client = API(connection, URL_PATH)

    r = await client.requests_get('/lookup/%s' % search)
    return await parse_response(r, LOOKUP_SCHEMA)


async def certifiers_of(connection, search):
    """
    GET UID/Public key certifiers

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str search: UID or public key
    :rtype: dict
    """

    client = API(connection, URL_PATH)

    r = await client.requests_get('/certifiers-of/%s' % search)
    return await parse_response(r, CERTIFICATIONS_SCHEMA)

async def certified_by(connection, search):
    """
    GET identities certified by UID/Public key

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str search: UID or public key
    :rtype: dict
    """
    client = API(connection, URL_PATH)

    r = await client.requests_get('/certified-by/%s' % search)
    return await parse_response(r, CERTIFICATIONS_SCHEMA)

async def members(connection):
    """
    GET list of all current members of the Web of Trust

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """
    client = API(connection, URL_PATH)

    r = await client.requests_get('/members')
    return await parse_response(r, MEMBERS_SCHEMA)


async def requirements(connection, search):
    """
    GET list of requirements for a given UID/Public key

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str search: UID or public key
    :rtype: dict
    """
    client = API(connection, URL_PATH)

    r = await client.requests_get('/requirements/%s' % search)
    return await parse_response(r, REQUIREMENTS_SCHEMA)



'''
Created on 13 déc. 2014

@author: inso
'''
import unittest
from duniterpy.documents.peer import Peer, BMAEndpoint, UnknownEndpoint


rawpeer = """Version: 2
Type: Peer
Currency: beta_brousouf
PublicKey: HsLShAtzXTVxeUtQd7yi5Z5Zh4zNvbu8sTEZ53nfKcqY
Block: 8-1922C324ABC4AF7EF7656734A31F5197888DDD52
Endpoints:
BASIC_MERKLED_API some.dns.name 88.77.66.55 2001:0db8:0000:85a3:0000:0000:ac1f 9001
BASIC_MERKLED_API some.dns.name 88.77.66.55 2001:0db8:0000:85a3:0000:0000:ac1f 9002
OTHER_PROTOCOL 88.77.66.55 9001
dkaXIiCYUJtCg8Feh/BKvPYf4uFH9CJ/zY6J4MlA9BsjmcMe4YAblvNt/gJy31b1aGq3ue3h14mLMCu84rraDg==
"""

test_weird_ipv6_peer = """Version: 10
Type: Peer
Currency: g1
PublicKey: 6fFt4zdvtNyVcfJn7Y41mKLmMDizyK3nVeNW3qdDXzpc
Block: 18198-000004AC710E04D8015ED6CA5D87D4B6620A7551233FFEE1B521FF756CE3B9CD
Endpoints:
BASIC_MERKLED_API duniter.aquilenet.fr 141.255.128.35 2a01:474::35 10901
BMAS duniter.aquilenet.fr 443
dkaXIiCYUJtCg8Feh/BKvPYf4uFH9CJ/zY6J4MlA9BsjmcMe4YAblvNt/gJy31b1aGq3ue3h14mLMCu84rraDg==
"""

class TestPeer(unittest.TestCase):
    def test_fromraw(self):
        peer = Peer.from_signed_raw(rawpeer)
        self.assertEqual(peer.currency, "beta_brousouf")
        self.assertEqual(peer.pubkey, "HsLShAtzXTVxeUtQd7yi5Z5Zh4zNvbu8sTEZ53nfKcqY")
        self.assertEqual(str(peer.blockUID), "8-1922C324ABC4AF7EF7656734A31F5197888DDD52")
        self.assertEqual(len(peer.endpoints), 3)
        self.assertTrue(type(peer.endpoints[0]) is BMAEndpoint)
        self.assertTrue(type(peer.endpoints[1]) is BMAEndpoint)
        self.assertTrue(type(peer.endpoints[2]) is UnknownEndpoint)

        self.assertEqual(peer.endpoints[0].server, "some.dns.name")
        self.assertEqual(peer.endpoints[0].ipv4, "88.77.66.55")
        self.assertEqual(peer.endpoints[0].ipv6, "2001:0db8:0000:85a3:0000:0000:ac1f")
        self.assertEqual(peer.endpoints[0].port, 9001)

        self.assertEqual(peer.endpoints[1].server, "some.dns.name")
        self.assertEqual(peer.endpoints[1].ipv4, "88.77.66.55")
        self.assertEqual(peer.endpoints[1].ipv6, "2001:0db8:0000:85a3:0000:0000:ac1f")
        self.assertEqual(peer.endpoints[1].port, 9002)

        self.assertEqual(peer.signatures[0], "dkaXIiCYUJtCg8Feh/BKvPYf4uFH9CJ/zY6J4MlA9BsjmcMe4YAblvNt/gJy31b1aGq3ue3h14mLMCu84rraDg==")

    def test_fromraw_toraw(self):
        peer = Peer.from_signed_raw(rawpeer)
        rendered_peer = peer.signed_raw()
        from_rendered_peer = Peer.from_signed_raw(rendered_peer)

        self.assertEqual(from_rendered_peer.currency, "beta_brousouf")
        self.assertEqual(from_rendered_peer.pubkey, "HsLShAtzXTVxeUtQd7yi5Z5Zh4zNvbu8sTEZ53nfKcqY")
        self.assertEqual(str(from_rendered_peer.blockUID), "8-1922C324ABC4AF7EF7656734A31F5197888DDD52")
        self.assertEqual(len(from_rendered_peer.endpoints), 3)
        self.assertTrue(type(from_rendered_peer.endpoints[0]) is BMAEndpoint)
        self.assertTrue(type(from_rendered_peer.endpoints[1]) is BMAEndpoint)
        self.assertTrue(type(from_rendered_peer.endpoints[2]) is UnknownEndpoint)

        self.assertEqual(from_rendered_peer.endpoints[0].server, "some.dns.name")
        self.assertEqual(from_rendered_peer.endpoints[0].ipv4, "88.77.66.55")
        self.assertEqual(from_rendered_peer.endpoints[0].ipv6, "2001:0db8:0000:85a3:0000:0000:ac1f")
        self.assertEqual(from_rendered_peer.endpoints[0].port, 9001)

        self.assertEqual(from_rendered_peer.endpoints[1].server, "some.dns.name")
        self.assertEqual(from_rendered_peer.endpoints[1].ipv4, "88.77.66.55")
        self.assertEqual(from_rendered_peer.endpoints[1].ipv6, "2001:0db8:0000:85a3:0000:0000:ac1f")
        self.assertEqual(from_rendered_peer.endpoints[1].port, 9002)

        self.assertEqual(from_rendered_peer.signatures[0], "dkaXIiCYUJtCg8Feh/BKvPYf4uFH9CJ/zY6J4MlA9BsjmcMe4YAblvNt/gJy31b1aGq3ue3h14mLMCu84rraDg==")
        self.assertEqual(rawpeer, from_rendered_peer.signed_raw())

    def test_incorrect(self):
        peer = Peer.from_signed_raw(test_weird_ipv6_peer)
        rendered_peer = peer.signed_raw()
        from_rendered_peer = Peer.from_signed_raw(rendered_peer)



import unittest
from duniterpy.api.bma import API, parse_error
from duniterpy.documents.peer import BMAEndpoint


class Test_BMA_API(unittest.TestCase):

    def test_reverse_url_complete(self):
        endpoint = BMAEndpoint("test.com", "124.2.2.1", "2001:0db8:0000:85a3:0000:0000:ac1f:8001 ", 9092)
        api = API(next(endpoint.conn_handler()), "any")
        self.assertEqual(api.reverse_url("http", "/test/url"), "http://test.com:9092/any/test/url")

    def test_reverse_url_only_ipv4(self):
        endpoint = BMAEndpoint(None, "124.2.2.1", None, 9092)
        api = API(next(endpoint.conn_handler()), "any")
        self.assertEqual(api.reverse_url("http", "/test/url"), "http://124.2.2.1:9092/any/test/url")

    def test_reverse_url_only_ipv6(self):
        endpoint = BMAEndpoint(None, None, "2001:0db8:0000:85a3:0000:0000:ac1f:8001", 9092)
        api = API(next(endpoint.conn_handler()), "any")
        self.assertEqual(api.reverse_url("http", "/test/url"), "http://[2001:0db8:0000:85a3:0000:0000:ac1f:8001]:9092/any/test/url")

    def test_parse_error(self):
        api = API(None, "any")
        error = parse_error("""{
"ucode": 1005,
"message": "Document has unkown fields or wrong line ending format"
}""")
        self.assertEqual(error["ucode"], 1005)
        self.assertEqual(error["message"], "Document has unkown fields or wrong line ending format")



#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Caner Candan <caner@candan.fr>, http://caner.candan.fr
#

MANAGED_API=["BASIC_MERKLED_API", "BMAS"]

__author__      = 'Caner Candan & inso & vit'
__version__     = '0.40.10'
__nonsense__    = 'duniter'

from . import api, documents, key


#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Caner Candan <caner@candan.fr>, http://caner.candan.fr
#

from duniterpy.api.bma import API, logging, parse_response

logger = logging.getLogger("duniter/blockchain")

URL_PATH = 'blockchain'

BLOCK_SCHEMA = {
    "type": "object",
    "properties": {
        "version": {
            "type": "number"
        },
        "currency": {
            "type": "string"
        },
        "nonce": {
            "type": "number"
        },
        "number": {
            "type": "number"
        },
        "time": {
            "type": "number"
        },
        "medianTime": {
            "type": "number"
        },
        "dividend": {
            "type": ["number", "null"]
        },
        "monetaryMass": {
            "type": ["number", "null"]
        },
        "issuer": {
            "type": "string"
        },
        "previousHash": {
            "type": ["string", "null"]
        },
        "previousIssuer": {
            "type": ["string", "null"]
        },
        "membersCount": {
            "type": "number"
        },
        "hash": {
            "type": "string"
        },
        "inner_hash": {
            "type": "string"
        },
        "identities": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "joiners": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "leavers": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "revoked": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "excluded": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "certifications": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "transactions": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "signatures": {
                        "type": "array"
                    },
                    "version": {
                        "type": "number"
                    },
                    "currency": {
                        "type": "string"
                    },
                    "issuers": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "inputs": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "unlocks": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "outputs": {
                        "type": "array",
                        "item": {
                            "type": "string"
                        }
                    }
                },
                "required": ["signatures", "version", "currency", "issuers", "inputs", "outputs"]
            }
        },
        "signature": {
            "type": "string"
        },
    },
    "required": ["version", "currency", "nonce", "number", "time", "medianTime", "dividend", "monetaryMass",
                 "issuer", "previousHash", "previousIssuer", "membersCount", "hash", "inner_hash", "identities",
                 "joiners", "leavers", "excluded", "certifications", "transactions", "signature"]
}

BLOCK_NUMBERS_SCHEMA = {
    "type": "object",
    "properties": {
        "result": {
            "type": "object",
            "properties": {
                "blocks": {
                    "type": "array",
                    "items": {
                        "type": "number"
                    }
                },
            },
            "required": ["blocks"]
        }
    },
    "required": ["result"]
}

PARAMETERS_SCHEMA = {
        "type": "object",
        "properties":
        {
              "currency": {
                  "type": "string"
              },
              "c": {
                  "type": "number"
              },
              "dt": {
                  "type": "number"
              },
              "ud0": {
                  "type": "number"
              },
              "sigPeriod": {
                  "type": "number"
              },
              "sigStock": {
                  "type": "number"
              },
              "sigWindow": {
                  "type": "number"
              },
              "sigValidity": {
                  "type": "number"
              },
              "sigQty": {
                  "type": "number"
              },
              "xpercent": {
                  "type": "number"
              },
              "msValidity": {
                  "type": "number"
              },
              "stepMax": {
                  "type": "number"
              },
              "medianTimeBlocks": {
                  "type": "number"
              },
              "avgGenTime": {
                  "type": "number"
              },
              "dtDiffEval": {
                  "type": "number"
              },
              "percentRot": {
                  "type": "number"
              },
              "udTime0": {
                  "type": "number"
              },
              "udReevalTime0": {
                  "type": "number"
              },
              "dtReeval": {
                  "type": "number"
              }
            },
        "required": ["currency", "c", "dt", "ud0","sigPeriod", "sigValidity", "sigQty", "xpercent", "sigStock",
                     "sigWindow", "msValidity","stepMax", "medianTimeBlocks",
                     "avgGenTime", "dtDiffEval", "percentRot", "udTime0", "udReevalTime0", "dtReeval"]
    }


MEMBERSHIPS_SCHEMA = {
        "type": "object",
        "properties":
        {
            "pubkey": {
                "type": "string"
            },
            "uid": {
                "type": "string",
            },
            "sigDate": {
                "type": "string"
            },
            "memberships": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "version": {
                            "type": "number"
                        },
                        "currency": {
                            "type": "string"
                        },
                        "membership": {
                            "type": "string"
                        },
                        "blockNumber": {
                            "type": "number"
                        },
                        "written": {
                            "type": ["number", "null"]
                        }
                    },
                    "required": ["version", "currency", "membership", "blockNumber", "blockHash", "written"]
                }
            }
        },
        "required": ["pubkey", "uid", "sigDate", "memberships"]
    }


BLOCKS_SCHEMA = {
    "type": "array",
    "items": BLOCK_SCHEMA
}

HARDSHIP_SCHEMA = {
    "type": "object",
    "properties": {
        "block": {
            "type": "number"
        },
        "level": {
            "type": "number"
        }
    },
    "required": ["block", "level"]
}


async def parameters(connection):
    """
    GET the blockchain parameters used by this node

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """
    client = API(connection, URL_PATH)
    r = await client.requests_get('/parameters')
    return await parse_response(r, PARAMETERS_SCHEMA)

async def memberships(connection, search):
    """
    GET list of Membership documents for UID/Public key

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str search: UID/Public key
    :rtype: dict
    """
    client = API(connection, URL_PATH)

    r = await client.requests_get('/memberships/%s' % search)
    return await parse_response(r, MEMBERSHIPS_SCHEMA)

async def membership(connection, membership):
    """
    POST a Membership document

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str membership: Membership signed raw document
    :rtype: aiohttp.ClientResponse
    """
    client = API(connection, URL_PATH)

    return await client.requests_post('/membership', membership=membership)

async def block(connection, number=0, block=None, signature=None):
    """
    GET/POST a block from/to the blockchain

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param int number: Block number to get
    :param dict block: Block document to post
    :param str signature: Signature of the block document issuer
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    # POST block
    if block is not None and signature is not None:
        return await client.requests_post('/block', block=block, signature=signature)

    # GET block
    r = await client.requests_get('/block/%d' % number)
    data = await parse_response(r, BLOCK_SCHEMA)
    return data

async def current(connection):
    """
    GET, return last accepted block

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """

    client = API(connection, URL_PATH)

    r = await client.requests_get('/current')
    return await parse_response(r, BLOCK_SCHEMA)


async def blocks(connection, count, start):
    """
    GET list of blocks from the blockchain

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param int count: Number of blocks
    :param int start: First block number
    :rtype: list
    """

    client = API(connection, URL_PATH)
    assert type(count) is int
    assert type(start) is int
    r = await client.requests_get('/blocks/%d/%d' % (count, start))
    return await parse_response(r, BLOCKS_SCHEMA)

async def hardship(connection, pubkey):
    """
    GET hardship level for given member's public key for writing next block

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str pubkey:  Public key of the member
    :rtype: dict
    """
    client = API(connection, URL_PATH)
    r = await client.requests_get('/hardship/%s' % pubkey)
    return await parse_response(r, HARDSHIP_SCHEMA)

async def newcomers(connection):
    """
    GET, return block numbers containing newcomers

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    r = await client.requests_get('/with/newcomers')
    return await parse_response(r, BLOCK_NUMBERS_SCHEMA)

async def certifications(connection):
    """
    GET, return block numbers containing certifications

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    r = await client.requests_get('/with/certs')
    return await parse_response(r, BLOCK_NUMBERS_SCHEMA)

async def joiners(connection):
    """
    GET, return block numbers containing joiners

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    r = await client.requests_get('/with/joiners')
    return await parse_response(r, BLOCK_NUMBERS_SCHEMA)

async def actives(connection):
    """
    GET, return block numbers containing actives

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    r = await client.requests_get('/with/actives')
    return await parse_response(r, BLOCK_NUMBERS_SCHEMA)

async def leavers(connection):
    """
    GET, return block numbers containing leavers

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    r = await client.requests_get('/with/leavers')
    return await parse_response(r, BLOCK_NUMBERS_SCHEMA)

async def excluded(connection):
    """
    GET, return block numbers containing excluded

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    r = await client.requests_get('/with/excluded')
    return await parse_response(r, BLOCK_NUMBERS_SCHEMA)

async def ud(connection):
    """
    GET, return block numbers containing universal dividend

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """
    client = API(connection, URL_PATH)
    r = await client.requests_get('/with/ud')
    return await parse_response(r, BLOCK_NUMBERS_SCHEMA)

async def tx(connection):
    """
    GET, return block numbers containing transactions

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    r = await client.requests_get('/with/tx')
    return await parse_response(r, BLOCK_NUMBERS_SCHEMA)



from .block import Block, BlockUID, block_uid
from .certification import Identity, Certification, Revocation
from .membership import Membership
from .peer import endpoint, BMAEndpoint, UnknownEndpoint, Peer, SecuredBMAEndpoint
from .transaction import SimpleTransaction, Transaction, InputSource, OutputSource, \
    SIGParameter, Unlock, UnlockParameter
from .document import Document, MalformedDocumentError

from . import constants



'''
Created on 6 déc. 2014

@author: inso
'''

import unittest
from duniterpy.documents.certification import Identity, Certification, Revocation
from duniterpy.documents import Block, BlockUID

selfcert_inlines = ["HnFcSms8jzwngtVomTTnzudZx7SHUQY8sVE1y8yBmULk:\
h/H8tDIEbfA4yxMQcvfOXVDQhi1sUa9qYtPKrM59Bulv97ouwbAvAsEkC1Uyit1IOpeAV+CQQs4IaAyjE8F1Cw==:\
32-DB30D958EE5CB75186972286ED3F4686B8A1C2CD:lolcat\n", "RdrHvL179Rw62UuyBrqy2M1crx7RPajaViBatS59EGS:\
Ah55O8cvdkGS4at6AGOKUjy+wrFwAq8iKRJ5xLIb6Xdi3M8WfGOUdMjwZA6GlSkdtlMgEhQPm+r2PMebxKrCBg==:\
36-1076F10A7397715D2BEE82579861999EA1F274AC:lolmouse\n" ]


cert_inlines = [
"8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU:HnFcSms8jzwngtVomTTnzudZx7SHUQY8sVE1y8yBmULk:\
0:TgmDuMxZdyutroj9jiLJA8tQp/389JIzDKuxW5+h7GIfjDu1ZbwI7HNm5rlUDhR2KreaV/QJjEaItT4Cf75rCQ==\n",
"9fx25FmeBDJcikZLWxK5HuzKNbY6MaWYXoK1ajteE42Y:8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU:12:\
qn/XNJjaGIwfnR+wGrDME6YviCQbG+ywsQWnETlAsL6q7o3k1UhpR5ZTVY9dvejLKuC+1mUEXVTmH+8Ib55DBA==\n"
]

revokation_inline = "8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU:\
TgmDuMxZdyutroj9jiLJA8tQp/389JIzDKuxW5+h7GIfjDu1ZbwI7HNm5rlUDhR2KreaV/QJjEaItT4Cf75rCQ==\n"


class Test_Certification(unittest.TestCase):

    def test_self_certification_from_inline(self):
        version = 2
        currency = "beta_brousouf"
        selfcert = Identity.from_inline(version, currency, selfcert_inlines[0])
        self.assertEqual(selfcert.pubkey, "HnFcSms8jzwngtVomTTnzudZx7SHUQY8sVE1y8yBmULk")
        self.assertEqual(selfcert.signatures[0], "h/H8tDIEbfA4yxMQcvfOXVDQhi1sUa9qYtPKrM59Bulv97ouwbAvAsEkC1Uyit1IOpeAV+CQQs4IaAyjE8F1Cw==")
        self.assertEqual(str(selfcert.timestamp), "32-DB30D958EE5CB75186972286ED3F4686B8A1C2CD")
        self.assertEqual(selfcert.uid, "lolcat")

        selfcert = Identity.from_inline(version, currency, selfcert_inlines[1])
        self.assertEqual(selfcert.pubkey, "RdrHvL179Rw62UuyBrqy2M1crx7RPajaViBatS59EGS")
        self.assertEqual(selfcert.signatures[0], "Ah55O8cvdkGS4at6AGOKUjy+wrFwAq8iKRJ5xLIb6Xdi3M8WfGOUdMjwZA6GlSkdtlMgEhQPm+r2PMebxKrCBg==")
        self.assertEqual(str(selfcert.timestamp), "36-1076F10A7397715D2BEE82579861999EA1F274AC")
        self.assertEqual(selfcert.uid, "lolmouse")

    def test_raw_self_certification(self):
        version = 2
        currency = "beta_brousouf"
        issuer = "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd"
        uid = "lolcat"
        timestamp = BlockUID(32, "DB30D958EE5CB75186972286ED3F4686B8A1C2CD")
        signature = "J3G9oM5AKYZNLAB5Wx499w61NuUoS57JVccTShUbGpCMjCqj9yXXqNq7dyZpDWA6BxipsiaMZhujMeBfCznzyci"

        selfcert = Identity(version, currency, issuer, uid, timestamp, signature)

        result = """Version: 2
Type: Identity
Currency: beta_brousouf
Issuer: HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd
UniqueID: lolcat
Timestamp: 32-DB30D958EE5CB75186972286ED3F4686B8A1C2CD
J3G9oM5AKYZNLAB5Wx499w61NuUoS57JVccTShUbGpCMjCqj9yXXqNq7dyZpDWA6BxipsiaMZhujMeBfCznzyci
"""
        self.assertEqual(selfcert.signed_raw(), result)

        from_raw = Identity.from_signed_raw(result)
        self.assertEqual(from_raw.signed_raw(), result)

    def test_certifications_from_inline(self):
        version = 2
        currency = "zeta_brousouf"
        cert = Certification.from_inline(version, currency, None, cert_inlines[0])
        self.assertEqual(cert.pubkey_from, "8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU")
        self.assertEqual(cert.pubkey_to, "HnFcSms8jzwngtVomTTnzudZx7SHUQY8sVE1y8yBmULk")
        self.assertEqual(cert.timestamp.number, 0)
        self.assertEqual(cert.timestamp.sha_hash, Block.Empty_Hash)
        self.assertEqual(cert.signatures[0], "TgmDuMxZdyutroj9jiLJA8tQp/389JIzDKuxW5+h7GIfjDu1ZbwI7HNm5rlUDhR2KreaV/QJjEaItT4Cf75rCQ==")

        cert = Certification.from_inline(version, currency, "DB30D958EE5CB75186972286ED3F4686B8A1C2CD", cert_inlines[1])
        self.assertEqual(cert.pubkey_from, "9fx25FmeBDJcikZLWxK5HuzKNbY6MaWYXoK1ajteE42Y")
        self.assertEqual(cert.pubkey_to, "8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU")
        self.assertEqual(cert.timestamp.number, 12)
        self.assertEqual(cert.timestamp.sha_hash, "DB30D958EE5CB75186972286ED3F4686B8A1C2CD")
        self.assertEqual(cert.signatures[0], "qn/XNJjaGIwfnR+wGrDME6YviCQbG+ywsQWnETlAsL6q7o3k1UhpR5ZTVY9dvejLKuC+1mUEXVTmH+8Ib55DBA==")

    def test_certification_raw(self):
        version = 2
        currency = "beta_brousouf"
        pubkey_from = "DNann1Lh55eZMEDXeYt59bzHbA3NJR46DeQYCS2qQdLV"
        pubkey_to = "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd"
        timestamp = BlockUID(36, "1076F10A7397715D2BEE82579861999EA1F274AC")
        signature = "SoKwoa8PFfCDJWZ6dNCv7XstezHcc2BbKiJgVDXv82R5zYR83nis9dShLgWJ5w48noVUHimdngzYQneNYSMV3rk"
        selfcert = Identity(version, currency, pubkey_to, "lolcat",
                            BlockUID(32, "DB30D958EE5CB75186972286ED3F4686B8A1C2CD"),
                                     "J3G9oM5AKYZNLAB5Wx499w61NuUoS57JVccTShUbGpCMjCqj9yXXqNq7dyZpDWA6BxipsiaMZhujMeBfCznzyci")

        certification = Certification(version, currency, pubkey_from, pubkey_to, timestamp, signature)

        result = """Version: 2
Type: Certification
Currency: beta_brousouf
Issuer: DNann1Lh55eZMEDXeYt59bzHbA3NJR46DeQYCS2qQdLV
IdtyIssuer: HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd
IdtyUniqueID: lolcat
IdtyTimestamp: 32-DB30D958EE5CB75186972286ED3F4686B8A1C2CD
IdtySignature: J3G9oM5AKYZNLAB5Wx499w61NuUoS57JVccTShUbGpCMjCqj9yXXqNq7dyZpDWA6BxipsiaMZhujMeBfCznzyci
CertTimestamp: 36-1076F10A7397715D2BEE82579861999EA1F274AC
SoKwoa8PFfCDJWZ6dNCv7XstezHcc2BbKiJgVDXv82R5zYR83nis9dShLgWJ5w48noVUHimdngzYQneNYSMV3rk
"""
        self.assertEqual(certification.signed_raw(selfcert), result)

        from_raw = Certification.from_signed_raw(certification.signed_raw(selfcert))
        self.assertEqual(from_raw.signed_raw(selfcert), result)

    def test_revokation_from_inline(self):
        version = 2
        currency = "zeta_brousouf"
        revokation = Revocation.from_inline(version, currency, revokation_inline)
        self.assertEqual(revokation.pubkey, "8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU")
        self.assertEqual(revokation.signatures[0], "TgmDuMxZdyutroj9jiLJA8tQp/389JIzDKuxW5+h7GIfjDu1ZbwI7HNm5rlUDhR2KreaV/QJjEaItT4Cf75rCQ==")

    def test_revokation_raw(self):

        version = 2
        currency = "beta_brousouf"
        pubkey = "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd"
        signature = "SoKwoa8PFfCDJWZ6dNCv7XstezHcc2BbKiJgVDXv82R5zYR83nis9dShLgWJ5w48noVUHimdngzYQneNYSMV3rk"
        revokation = Revocation(version, currency, pubkey, signature)
        selfcert = Identity(version, currency, pubkey, "lolcat",
                            BlockUID(32, "DB30D958EE5CB75186972286ED3F4686B8A1C2CD"),
                                     "J3G9oM5AKYZNLAB5Wx499w61NuUoS57JVccTShUbGpCMjCqj9yXXqNq7dyZpDWA6BxipsiaMZhujMeBfCznzyci")

        result = """Version: 2
Type: Revocation
Currency: beta_brousouf
Issuer: HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd
IdtyUniqueID: lolcat
IdtyTimestamp: 32-DB30D958EE5CB75186972286ED3F4686B8A1C2CD
IdtySignature: J3G9oM5AKYZNLAB5Wx499w61NuUoS57JVccTShUbGpCMjCqj9yXXqNq7dyZpDWA6BxipsiaMZhujMeBfCznzyci
SoKwoa8PFfCDJWZ6dNCv7XstezHcc2BbKiJgVDXv82R5zYR83nis9dShLgWJ5w48noVUHimdngzYQneNYSMV3rk
"""
        self.assertEqual(revokation.signed_raw(selfcert), result)

    def test_revokation_from_signed_raw(self):
        signed_raw = """Version: 2
Type: Revocation
Currency: beta_brousouf
Issuer: HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd
IdtyUniqueID: lolcat
IdtyTimestamp: 32-DB30D958EE5CB75186972286ED3F4686B8A1C2CD
IdtySignature: J3G9oM5AKYZNLAB5Wx499w61NuUoS57JVccTShUbGpCMjCqj9yXXqNq7dyZpDWA6BxipsiaMZhujMeBfCznzyci
SoKwoa8PFfCDJWZ6dNCv7XstezHcc2BbKiJgVDXv82R5zYR83nis9dShLgWJ5w48noVUHimdngzYQneNYSMV3rk
"""
        revocation = Revocation.from_signed_raw(signed_raw)
        selfcert = Revocation.extract_self_cert(signed_raw)
        self.assertEqual(revocation.signed_raw(selfcert), signed_raw)



from duniterpy.key import VerifyingKey, SigningKey, ScryptParams
from duniterpy.documents import Peer
import unittest


class TestVerifyingKey(unittest.TestCase):
    def test_from_sign_to_verify(self):
        sign_key = SigningKey("saltsalt", "passwordpassword", ScryptParams(4096, 16, 1))
        verify_key = VerifyingKey(sign_key.pubkey)
        self.assertEqual(verify_key.vk, sign_key.vk)

    def test_peer_signature(self):
        signed_raw = """Version: 2
Type: Peer
Currency: test_net
PublicKey: 8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU
Block: 2086-00005D6FC6E22FB308D8815A565A01C66FFB7DC761D616DE0698F6322565F1D6
Endpoints:
BASIC_MERKLED_API testnet.duniter.inso.ovh 80
4aQ/sfqFAFUeYkkLdC2OfgXqTBjCIcMptpR/GIlGqbe4aFVJcy9NEVAFx7sHiLuAb+VNnec3XHHC+xOk3MLzDA==
"""""
        peer = Peer.from_signed_raw(signed_raw)
        pubkey = "8Fi1VSTbjkXguwThF4v2ZxC5whK7pwG2vcGTkPUPjPGU"
        verifying_key = VerifyingKey(pubkey)
        self.assertTrue(verifying_key.verify_document(peer))



import asyncio
import socket
from aiohttp import web
from aiohttp import log


# Thanks to aiohttp tests courtesy
# Here is a nice mocking server
class WebFunctionalSetupMixin:

    def setUp(self):
        self.handler = None
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        if self.handler:
            self.loop.run_until_complete(self.handler.finish_connections())
        try:
            self.loop.stop()
            self.loop.close()
        finally:
            asyncio.set_event_loop(None)

    def find_unused_port(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(('127.0.0.1', 0))
        port = s.getsockname()[1]
        s.close()
        return port

    async    def create_server(self, method, path, handler=None, ssl_ctx=None):
        app = web.Application(loop=self.loop)
        if handler:
            app.router.add_route(method, path, handler)

        port = self.find_unused_port()
        self.handler = app.make_handler(
            keep_alive_on=False,
            access_log=log.access_logger)
        srv = await self.loop.create_server(
            self.handler, '127.0.0.1', port, ssl=ssl_ctx)
        protocol = "https" if ssl_ctx else "http"
        url = "{}://127.0.0.1:{}".format(protocol, port) + path
        self.addCleanup(srv.close)
        return app, srv, port, url


#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Caner Candan <caner@candan.fr>, http://caner.candan.fr
#

from duniterpy.api.bma import API, logging
from duniterpy.api.bma.blockchain import BLOCK_SCHEMA

logger = logging.getLogger("duniter/ws")

URL_PATH = 'ws'


WS_BLOCk_SCHEMA = BLOCK_SCHEMA
WS_PEER_SCHEMA = {
    "type": "object",
    "properties": {
        "version": {
            "type": "number"
        },
        "currency": {
            "type": "string"
        },
        "pubkey": {
            "type": "string"
        },
        "endpoints": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "signature": {
            "type": "string"
        }
    },
    "required": ["version", "currency", "pubkey", "endpoints", "signature"]
}


def block(connection):
    """
    Connect to block websocket

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: aiohttp.ClientWebSocketResponse
    """
    client = API(connection, URL_PATH)
    return client.connect_ws('/block')


def peer(connection):
    """
    Connect to peer websocket

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: aiohttp.ClientWebSocketResponse
    """
    client = API(connection, URL_PATH)
    return client.connect_ws('/peer')



#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Caner Candan <caner@candan.fr>, http://caner.candan.fr
# Inso <insomniak.fr at gmail.com>


import aiohttp
import json
import logging
import jsonschema
from ..errors import DuniterError

logger = logging.getLogger("duniter")

ERROR_SCHEMA = {
    "type": "object",
    "properties": {
        "ucode": {
            "type": "number"
        },
        "message": {
            "type": "string"
        }
    },
    "required": ["ucode", "message"]
}


class ConnectionHandler(object):
    """Helper class used by other API classes to ease passing server connection information."""

    def __init__(self, http_scheme, ws_scheme, server, port, path="", proxy=None, session=None):
        """
        Init instance of connection handler

        :param str server: Server IP or domaine name
        :param int port: Port
        :param aiohttp.ClientSession|None session: Session AIOHTTP
        """
        self.http_scheme = http_scheme
        self.ws_scheme = ws_scheme
        self.server = server
        self.proxy = proxy
        self.port = port
        self.session = session
        self.path = path

    def __str__(self):
        return 'connection info: %s:%d' % (self.server, self.port)


def parse_text(text, schema):
    """
    Validate and parse the BMA answer from websocket

    :param str text: the bma answer
    :return: the json data
    """
    try:
        data = json.loads(text)
        jsonschema.validate(data, schema)
        return data
    except (TypeError, json.decoder.JSONDecodeError):
        raise jsonschema.ValidationError("Could not parse json")


def parse_error(text):
    """
    Validate and parse the BMA answer from websocket

    :param str text: the bma error
    :return: the json data
    """
    try:
        data = json.loads(text)
        jsonschema.validate(data, ERROR_SCHEMA)
        return data
    except (TypeError, json.decoder.JSONDecodeError) as e:
        raise jsonschema.ValidationError("Could not parse json : {0}".format(str(e)))


async def parse_response(response, schema):
    """
    Validate and parse the BMA answer

    :param aiohttp.ClientResponse response: Response of aiohttp request
    :param dict schema: The expected response structure
    :return: the json data
    """
    try:
        data = await response.json()
        response.close()
        if schema is not None:
            jsonschema.validate(data, schema)
        return data
    except (TypeError, json.decoder.JSONDecodeError) as e:
        raise jsonschema.ValidationError("Could not parse json : {0}".format(str(e)))


class API(object):
    """APIRequest is a class used as an interface. The intermediate derivated classes are the modules and the leaf classes are the API requests."""

    schema = {}

    def __init__(self, connection_handler, module):
        """
        Asks a module in order to create the url used then by derivated classes.

        :param ConnectionHandler connection_handler: Connection handler
        :param str module: Module path
        """
        self.module = module
        self.connection_handler = connection_handler
        self.headers = {}

    def reverse_url(self, scheme, path):
        """
        Reverses the url using scheme and path given in parameter.

        :param str scheme: Scheme of the url
        :param str path: Path of the url
        :return: str
        """

        server, port = self.connection_handler.server, self.connection_handler.port
        if self.connection_handler.path:
            url = '{scheme}://{server}:{port}/{path}/{module}'.format(scheme=scheme,
                                                                  server=server,
                                                                  port=port,
                                                                  path=path,
                                                                  module=self.module)
        else:
            url = '{scheme}://{server}:{port}/{module}'.format(scheme=scheme,
                                                                  server=server,
                                                                  port=port,
                                                                  module=self.module)

        return url + path

    async def requests_get(self, path, **kwargs):
        """
        Requests GET wrapper in order to use API parameters.

        :param str path: the request path
        :rtype: aiohttp.ClientResponse
        """
        logging.debug("Request : {0}".format(self.reverse_url(self.connection_handler.http_scheme, path)))
        url = self.reverse_url(self.connection_handler.http_scheme, path)
        response = await self.connection_handler.session.get(url, params=kwargs, headers=self.headers,
                                                             proxy=self.connection_handler.proxy,
                                                             timeout=15)
        if response.status != 200:
            try:
                error_data = parse_error(await response.text())
                raise DuniterError(error_data)
            except (TypeError, jsonschema.ValidationError):
                raise ValueError('status code != 200 => %d (%s)' % (response.status, (await response.text())))

        return response

    async def requests_post(self, path, **kwargs):
        """
        Requests POST wrapper in order to use API parameters.

        :param str path: the request path
        :rtype: aiohttp.ClientResponse
        """
        if 'self_' in kwargs:
            kwargs['self'] = kwargs.pop('self_')

        logging.debug("POST : {0}".format(kwargs))
        with aiohttp.Timeout(15):
            response = await self.connection_handler.session.post(
                self.reverse_url(self.connection_handler.http_scheme, path),
                data=kwargs,
                headers=self.headers,
                proxy=self.connection_handler.proxy
            )
            return response

    def connect_ws(self, path):
        """
        Connect to a websocket in order to use API parameters

        :param str path: the url path
        :rtype: aiohttp.ClientWebSocketResponse
        """
        url = self.reverse_url(self.connection_handler.ws_scheme, path)
        return self.connection_handler.session.ws_connect(url, proxy=self.connection_handler.proxy)



#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Caner Candan <caner@candan.fr>, http://caner.candan.fr
#

from duniterpy.api.bma import logging, API, parse_response

logger = logging.getLogger("duniter/ud")

URL_PATH = 'ud'

UD_SCHEMA = {
    "type": "object",
    "properties": {
        "currency": {
            "type": "string"
        },
        "pubkey": {
            "type": "string"
        },
        "history": {
            "type": "object",
            "properties": {
                "history": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "block_number": {
                                "type": "number"
                            },

                            "consumed": {
                                "type": "boolean"
                            },
                            "time": {
                                "type": "number"
                            },
                            "amount": {
                                "type": "number"
                            },
                            "base": {
                                "type": "number"
                            },
                        }
                    }
                }
            }
        }
    },
    "required": ["currency", "pubkey", "history"]
}

async def history(connection, pubkey):
    """
    Get UD history of a member account

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param str pubkey:  Public key of the member

    :rtype: dict
    """
    client = API(connection, URL_PATH)

    r = await client.requests_get('/history/%s' % pubkey)
    return await parse_response(r, UD_SCHEMA)



import base58

class Base58Encoder(object):
    @staticmethod
    def encode(data):
        return base58.b58encode(data)

    @staticmethod
    def decode(data):
        return base58.b58decode(data)


from .document import Document, MalformedDocumentError
from .constants import pubkey_regex, transaction_hash_regex, block_id_regex, block_uid_regex, conditions_regex
from ..grammars import output
import pypeg2
import re


def reduce_base(amount, base):
    """
    Compute the reduced base of the given parameters
    :param int amount: the amount value
    :param int base: current base value
    :return: tuple containing computed (amount, base)
    :rtype: tuple
    """
    if amount == 0:
        return 0, 0

    next_amount = amount
    next_base = base
    while int(next_amount) == next_amount:
        amount = next_amount
        base = next_base
        next_amount /= 10
        next_base += 1
    return int(amount), int(base)


class Transaction(Document):
    """
.. note:: A transaction document is specified by the following format :

    | Document format :
    | Version: VERSION
    | Type: Transaction
    | Currency: CURRENCY_NAME
    | Issuers:
    | PUBLIC_KEY
    | ...
    | Inputs:
    | INDEX:SOURCE:NUMBER:FINGERPRINT:AMOUNT
    | ...
    | Outputs:
    | PUBLIC_KEY:AMOUNT
    | ...
    | Comment: COMMENT
    | ...
    |
    |
    | Compact format :
    | TX:VERSION:NB_ISSUERS:NB_INPUTS:NB_OUTPUTS:HAS_COMMENT
    | PUBLIC_KEY:INDEX
    | ...
    | INDEX:SOURCE:FINGERPRINT:AMOUNT
    | ...
    | PUBLIC_KEY:AMOUNT
    | ...
    | COMMENT
    | SIGNATURE
    | ...

    """

    re_type = re.compile("Type: (Transaction)\n")
    re_header = re.compile("TX:([0-9]+):([0-9]+):([0-9]+):([0-9]+):([0-9]+):(0|1):([0-9]+)\n")
    re_compact_blockstamp = re.compile("({block_uid_regex})\n".format(block_uid_regex=block_uid_regex))
    re_blockstamp = re.compile("Blockstamp: ({block_uid_regex})\n".format(block_uid_regex=block_uid_regex))
    re_locktime = re.compile("Locktime: ([0-9]+)\n")
    re_issuers = re.compile("Issuers:\n")
    re_inputs = re.compile("Inputs:\n")
    re_unlocks = re.compile("Unlocks:\n")
    re_outputs = re.compile("Outputs:\n")
    re_compact_comment = re.compile("([^\n]+)\n")
    re_comment = re.compile("Comment: ([^\n]*)\n")
    re_pubkey = re.compile("({pubkey_regex})\n".format(pubkey_regex=pubkey_regex))

    fields_parsers = {**Document.fields_parsers, **{
            "Type": re_type,
            "Blockstamp": re_blockstamp,
            "CompactBlockstamp": re_compact_blockstamp,
            "Locktime": re_locktime,
            "TX": re_header,
            "Issuers": re_issuers,
            "Inputs": re_inputs,
            "Unlocks": re_unlocks,
            "Outputs": re_outputs,
            "Comment": re_comment,
            "Compact comment": re_compact_comment,
            "Pubkey": re_pubkey
        }
    }

    def __init__(self, version, currency, blockstamp, locktime, issuers, inputs, unlocks, outputs,
                 comment, signatures):
        """

        :param int version:
        :param str currency:
        :param BlockUID blockstamp:
        :param int locktime:
        :param list[str] issuers:
        :param list[InputSource] inputs:
        :param list[Unlock] unlocks:
        :param list[OutputSource] outputs:
        :param comment:
        :param signatures:
        """
        super().__init__(version, currency, signatures)
        self.blockstamp = blockstamp
        self.locktime = locktime
        self.issuers = issuers
        self.inputs = inputs
        self.unlocks = unlocks
        self.outputs = outputs
        self.comment = comment

    @classmethod
    def from_bma_history(cls, currency, tx_data):
        """
        Get the transaction from json
        :param str currency: the currency of the tx
        :param dict tx_data: json data of the transaction
        :rtype: Transaction
        """
        tx_data = tx_data.copy()
        tx_data["currency"] = currency
        for data_list in ('issuers', 'outputs', 'inputs', 'unlocks', 'signatures'):
            tx_data['multiline_{0}'.format(data_list)] = '\n'.join(tx_data[data_list])
        if tx_data["version"] >= 3:
                signed_raw = """Version: {version}
Type: Transaction
Currency: {currency}
Blockstamp: {blockstamp}
Locktime: {locktime}
Issuers:
{multiline_issuers}
Inputs:
{multiline_inputs}
Unlocks:
{multiline_unlocks}
Outputs:
{multiline_outputs}
Comment: {comment}
{multiline_signatures}
""".format(**tx_data)
        else:
            signed_raw = """Version: {version}
Type: Transaction
Currency: {currency}
Locktime: {locktime}
Issuers:
{multiline_issuers}
Inputs:
{multiline_inputs}
Unlocks:
{multiline_unlocks}
Outputs:
{multiline_outputs}
Comment: {comment}
{multiline_signatures}
""".format(**tx_data)
        return Transaction.from_signed_raw(signed_raw)

    @classmethod
    def from_compact(cls, currency, compact):
        from .block import BlockUID
        lines = compact.splitlines(True)
        n = 0

        header_data = Transaction.re_header.match(lines[n])
        if header_data is None:
            raise MalformedDocumentError("Compact TX header")
        version = int(header_data.group(1))
        issuers_num = int(header_data.group(2))
        inputs_num = int(header_data.group(3))
        unlocks_num = int(header_data.group(4))
        outputs_num = int(header_data.group(5))
        has_comment = int(header_data.group(6))
        locktime = int(header_data.group(7))
        n += 1

        if version >= 3:
            blockstamp = BlockUID.from_str(Transaction.parse_field("CompactBlockstamp", lines[n]))
            n += 1
        else:
            blockstamp = None

        issuers = []
        inputs = []
        unlocks = []
        outputs = []
        signatures = []
        for i in range(0, issuers_num):
            issuer = Transaction.parse_field("Pubkey", lines[n])
            issuers.append(issuer)
            n += 1

        for i in range(0, inputs_num):
            input_source = InputSource.from_inline(version, lines[n])
            inputs.append(input_source)
            n += 1

        for i in range(0, unlocks_num):
            unlock = Unlock.from_inline(lines[n])
            unlocks.append(unlock)
            n += 1

        for i in range(0, outputs_num):
            output_source = OutputSource.from_inline(lines[n])
            outputs.append(output_source)
            n += 1

        comment = ""
        if has_comment == 1:
            if Transaction.re_compact_comment.match(lines[n]):
                comment = Transaction.re_compact_comment.match(lines[n]).group(1)
                n += 1
            else:
                raise MalformedDocumentError("Compact TX Comment")

        while n < len(lines):
            if Transaction.re_signature.match(lines[n]):
                signatures.append(Transaction.re_signature.match(lines[n]).group(1))
                n += 1
            else:
                raise MalformedDocumentError("Compact TX Signatures")

        return cls(version, currency, blockstamp, locktime, issuers, inputs, unlocks, outputs, comment, signatures)

    @classmethod
    def from_signed_raw(cls, raw):
        from .block import BlockUID
        lines = raw.splitlines(True)
        n = 0

        version = int(Transaction.parse_field("Version", lines[n]))
        n += 1

        Transaction.parse_field("Type", lines[n])
        n += 1

        currency = Transaction.parse_field("Currency", lines[n])
        n += 1

        if version >= 3:
            blockstamp = BlockUID.from_str(Transaction.parse_field("Blockstamp", lines[n]))
            n += 1
        else:
            blockstamp = None

        locktime = Transaction.parse_field("Locktime", lines[n])
        n += 1

        issuers = []
        inputs = []
        unlocks = []
        outputs = []
        signatures = []

        if Transaction.re_issuers.match(lines[n]):
            n += 1
            while Transaction.re_inputs.match(lines[n]) is None:
                issuer = Transaction.parse_field("Pubkey", lines[n])
                issuers.append(issuer)
                n += 1

        if Transaction.re_inputs.match(lines[n]):
            n += 1
            while Transaction.re_unlocks.match(lines[n]) is None:
                input_source = InputSource.from_inline(version, lines[n])
                inputs.append(input_source)
                n += 1

        if Transaction.re_unlocks.match(lines[n]):
            n += 1
            while Transaction.re_outputs.match(lines[n]) is None:
                unlock = Unlock.from_inline(lines[n])
                unlocks.append(unlock)
                n += 1

        if Transaction.re_outputs.match(lines[n]) is not None:
            n += 1
            while not Transaction.re_comment.match(lines[n]):
                output = OutputSource.from_inline(lines[n])
                outputs.append(output)
                n += 1

        comment = Transaction.parse_field("Comment", lines[n])
        n += 1

        if Transaction.re_signature.match(lines[n]) is not None:
            while n < len(lines):
                sign = Transaction.parse_field("Signature", lines[n])
                signatures.append(sign)
                n += 1

        return cls(version, currency, blockstamp, locktime, issuers, inputs, unlocks, outputs,
                   comment, signatures)

    def raw(self):
        doc = """Version: {0}
Type: Transaction
Currency: {1}
""".format(self.version,
           self.currency)

        if self.version >= 3:
            doc += "Blockstamp: {0}\n".format(self.blockstamp)

        doc += "Locktime: {0}\n".format(self.locktime)

        doc += "Issuers:\n"
        for p in self.issuers:
            doc += "{0}\n".format(p)

        doc += "Inputs:\n"
        for i in self.inputs:
            doc += "{0}\n".format(i.inline(self.version))

        doc += "Unlocks:\n"
        for u in self.unlocks:
            doc += "{0}\n".format(u.inline())

        doc += "Outputs:\n"
        for o in self.outputs:
            doc += "{0}\n".format(o.inline())

        doc += "Comment: "
        doc += "{0}\n".format(self.comment)

        return doc

    def compact(self):
        """
        Return a transaction in its compact format.
        """
        """TX:VERSION:NB_ISSUERS:NB_INPUTS:NB_UNLOCKS:NB_OUTPUTS:HAS_COMMENT:LOCKTIME
PUBLIC_KEY:INDEX
...
INDEX:SOURCE:FINGERPRINT:AMOUNT
...
PUBLIC_KEY:AMOUNT
...
COMMENT
"""
        doc = "TX:{0}:{1}:{2}:{3}:{4}:{5}:{6}\n".format(self.version,
                                              len(self.issuers),
                                              len(self.inputs),
                                              len(self.unlocks),
                                              len(self.outputs),
                                              '1' if self.comment != "" else '0',
                                               self.locktime)
        if self.version >= 3:
            doc += "{0}\n".format(self.blockstamp)

        for pubkey in self.issuers:
            doc += "{0}\n".format(pubkey)
        for i in self.inputs:
            doc += "{0}\n".format(i.inline(self.version))
        for u in self.unlocks:
            doc += "{0}\n".format(u.inline())
        for o in self.outputs:
            doc += "{0}\n".format(o.inline())
        if self.comment != "":
            doc += "{0}\n".format(self.comment)
        for s in self.signatures:
            doc += "{0}\n".format(s)

        return doc


class SimpleTransaction(Transaction):
    """
    As transaction class, but for only one issuer.
    ...
    """
    def __init__(self, version, currency, issuer,
                 single_input, outputs, comment, signature):
        """
        Constructor
        """
        super().__init__(version, currency, [issuer], [single_input],
              outputs, comment, [signature])

    @staticmethod
    def is_simple(tx):
        """
        Filter a transaction and checks if it is a basic one
        A simple transaction is a tx which has only one issuer
        and two outputs maximum. The unlocks must be done with
        simple "SIG" functions, and the outputs must be simple
        SIG conditions.
        :param duniterpy.documents.Transaction tx: the transaction to check
        :return: True if a simple transaction
        """
        simple = True
        if len(tx.issuers) != 1:
            simple = False
        for unlock in tx.unlocks:
            if len(unlock.parameters) != 1:
                simple = False
            elif type(unlock.parameters[0]) is not SIGParameter:
                simple = False
        for o in tx.outputs:
            if getattr('right', o.conditions, None):
                simple = False
            elif type(o.conditions.left) is not output.SIG:
                simple = False
        return simple


class InputSource:
    """
    A Transaction INPUT

.. note:: Compact :
    INDEX:SOURCE:FINGERPRINT:AMOUNT

    """
    re_inline = re.compile("(?:(?:(D):({pubkey_regex}):({block_id_regex}))|(?:(T):({transaction_hash_regex}):([0-9]+)))\n"
                           .format(pubkey_regex=pubkey_regex,
                                   block_id_regex=block_id_regex,
                                    transaction_hash_regex=transaction_hash_regex))
    re_inline_v3 = re.compile("([0-9]+):([0-9]+):(?:(?:(D):({pubkey_regex}):({block_id_regex}))|(?:(T):({transaction_hash_regex}):([0-9]+)))\n"
                           .format(pubkey_regex=pubkey_regex,
                                   block_id_regex=block_id_regex,
                                    transaction_hash_regex=transaction_hash_regex))

    def __init__(self, amount, base, source, origin_id, index):
        """
        An input source can come from a dividend or a transaction.

        :param int amount: amount of the input
        :param int base: base of the input
        :param str source: D if dividend, T if transaction
        :param str origin_id: a Public key if a dividend, a tx hash if a transaction
        :param int index: a block id if a dividend, an tx index if a transaction
        :return:
        """
        self.amount = amount
        self.base = base
        self.source = source
        self.origin_id = origin_id
        self.index = index

    @classmethod
    def from_inline(cls, tx_version, inline):
        if tx_version == 2:
            data = InputSource.re_inline.match(inline)
            if data is None:
                raise MalformedDocumentError("Inline input")
            source_offset = 0
            amount = 0
            base = 0
        else:
            data = InputSource.re_inline_v3.match(inline)
            if data is None:
                raise MalformedDocumentError("Inline input")
            source_offset = 2
            amount = data.group(1)
            base = data.group(2)
        if data.group(1 + source_offset):
            source = data.group(1 + source_offset)
            origin_id = data.group(2 + source_offset)
            index = int(data.group(3 + source_offset))
        else:
            source = data.group(4 + source_offset)
            origin_id = data.group(5 + source_offset)
            index = int(data.group(6 + source_offset))

        return cls(amount, base, source, origin_id, index)

    def inline(self, tx_version):
        if tx_version == 2:
            return "{0}:{1}:{2}".format(self.source,
                                        self.origin_id,
                                        self.index)
        else:
            return "{0}:{1}:{2}:{3}:{4}".format(self.amount,
                                        self.base,
                                        self.source,
                                        self.origin_id,
                                        self.index)


class UnlockParameter:

    def __init__(self):
        pass

    @classmethod
    def from_parameter(cls, parameter):
        for params_type in (SIGParameter, XHXParameter):
            param = params_type.from_parameter(parameter)
            if param:
                return param

    def compute(self):
        pass


class SIGParameter:
    """
    A Transaction UNLOCK SIG parameter
    """
    re_sig = re.compile("SIG\(([0-9]+)\)")

    def __init__(self, index):
        self.index = index

    @classmethod
    def from_parameter(cls, parameter):
        sig = SIGParameter.re_sig.match(parameter)
        if sig:
            return SIGParameter(sig.group(1))
        else:
            return None

    def __str__(self):
        return "SIG({0})".format(self.index)


class XHXParameter:
    """
    A Transaction UNLOCK XHX parameter
    """
    re_xhx = re.compile("XHX\(([0-9]+)\)")

    def __init__(self, integer):
        self.integer = integer

    @classmethod
    def from_parameter(cls, parameter):
        xhx = XHXParameter.re_xhx.match(parameter)
        if xhx:
            return XHXParameter(xhx.group(1))
        else:
            return None

    def compute(self):
        return

    def __str__(self):
        return "XHX({0})".format(self.integer)


class Unlock:
    """
    A Transaction UNLOCK
    """
    re_inline = re.compile("([0-9]+):((?:SIG\([0-9]+\)|XHX\([0-9]+\)|\s)+)\n")

    def __init__(self, index, parameters):
        self.index = index
        self.parameters = parameters

    @classmethod
    def from_inline(cls, inline):
        data = Unlock.re_inline.match(inline)
        if data is None:
            raise MalformedDocumentError("Inline input")
        index = int(data.group(1))
        parameters_str = data.group(2).split(' ')
        parameters = []
        for p in parameters_str:
            param = UnlockParameter.from_parameter(p)
            if param:
                parameters.append(param)
        return cls(index, parameters)

    def inline(self):
        return "{0}:{1}".format(self.index, ' '.join([str(p) for p in self.parameters]))


class OutputSource:
    """
    A Transaction OUTPUT
    """
    re_inline = re.compile("([0-9]+):([0-9]+):(.*)\n")

    def __init__(self, amount, base, conditions):
        self.amount = amount
        self.base = base
        self.conditions = conditions

    @classmethod
    def from_inline(cls, inline):
        data = OutputSource.re_inline.match(inline)
        if data is None:
            raise MalformedDocumentError("Inline output")
        amount = int(data.group(1))
        base = int(data.group(2))
        conditions_text = data.group(3)
        try:
            conditions = pypeg2.parse(conditions_text, output.Condition)
        except SyntaxError:
            raise MalformedDocumentError("Output source syntax error")
        return cls(amount, base, conditions)

    def inline(self):
        return "{0}:{1}:{2}".format(self.amount, self.base,
                                    pypeg2.compose(self.conditions, output.Condition))



"""
duniter public and private keys

@author: inso
"""

import libnacl.public
from pylibscrypt import scrypt
from .base58 import Base58Encoder
from .signing_key import _ensure_bytes


SEED_LENGTH = 32  # Length of the key
crypto_sign_BYTES = 64
SCRYPT_PARAMS = {'N': 4096,
                 'r': 16,
                 'p': 1
                 }


class SecretKey(libnacl.public.SecretKey):
    def __init__(self, salt, password):
        salt = _ensure_bytes(salt)
        password = _ensure_bytes(password)
        seed = scrypt(password, salt,
                    SCRYPT_PARAMS['N'], SCRYPT_PARAMS['r'], SCRYPT_PARAMS['p'],
                    SEED_LENGTH)

        super().__init__(seed)
        self.public_key = PublicKey(Base58Encoder.encode(self.pk))

    def encrypt(self, pubkey, noonce, text):
        text_bytes = _ensure_bytes(text)
        noonce_bytes = _ensure_bytes(noonce)
        recipient_pubkey = PublicKey(pubkey)
        crypt_bytes = libnacl.public.Box(self, recipient_pubkey).encrypt(text_bytes, noonce_bytes)
        return Base58Encoder.encode(crypt_bytes[24:])

    def decrypt(self, pubkey, noonce, text):
        sender_pubkey = PublicKey(pubkey)
        noonce_bytes = _ensure_bytes(noonce)
        encrypt_bytes = Base58Encoder.decode(text)
        decrypt_bytes = libnacl.public.Box(self, sender_pubkey).decrypt(encrypt_bytes, noonce_bytes)
        return decrypt_bytes.decode('utf-8')


class PublicKey(libnacl.public.PublicKey):
    def __init__(self, pubkey):
        key = Base58Encoder.decode(pubkey)
        super().__init__(key)

    def base58(self):
        return Base58Encoder.encode(self.pk)




from duniterpy.grammars import output
import unittest
import pypeg2


class Test_OutputGrammar(unittest.TestCase):
    def test_sig(self):
        condition = "SIG(HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd)"
        result = pypeg2.parse(condition, output.SIG)
        self.assertEqual(result.pubkey, "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd")
        self.assertEqual(pypeg2.compose(result, output.Condition), condition)

    def test_xhx(self):
        condition = "XHX(309BC5E644F797F53E5A2065EAF38A173437F2E6)"
        result = pypeg2.parse(condition, output.XHX)
        self.assertEqual(result.sha_hash, "309BC5E644F797F53E5A2065EAF38A173437F2E6")

    def test_sig_condition(self):
        condition = "SIG(HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd)"
        result = pypeg2.parse(condition, output.Condition)
        self.assertEqual(result.left.pubkey, "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd")
        self.assertEqual(pypeg2.compose(result, output.Condition), condition)

    def test_xhr_condition(self):
        condition = "SIG(HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd)"
        result = pypeg2.parse(condition, output.Condition)
        self.assertEqual(result.left.pubkey, "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd")
        self.assertEqual(pypeg2.compose(result, output.Condition), condition)

    def test_simple_condition(self):
        condition = "(SIG(HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd))"
        result = pypeg2.parse(condition, output.Condition)
        self.assertEqual(result.left.left.pubkey, "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd")
        self.assertEqual(pypeg2.compose(result, output.Condition), condition)

    def test_simple_and_condition(self):
        condition = "(SIG(HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd) && XHX(309BC5E644F797F53E5A2065EAF38A173437F2E6))"
        result = pypeg2.parse(condition, output.Condition)
        self.assertEqual(result.left.left.pubkey, "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd")
        self.assertEqual(result.left.op.name, "&&")
        self.assertEqual(result.left.right.sha_hash, "309BC5E644F797F53E5A2065EAF38A173437F2E6")
        self.assertEqual(pypeg2.compose(result, output.Condition), condition)

    def test_simple_or_condition(self):
        condition = "(SIG(HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd) || XHX(309BC5E644F797F53E5A2065EAF38A173437F2E6))"
        result = pypeg2.parse(condition, output.Condition)
        self.assertEqual(result.left.left.pubkey, "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd")
        self.assertEqual(result.left.op.name, "||")
        self.assertEqual(result.left.right.sha_hash, "309BC5E644F797F53E5A2065EAF38A173437F2E6")
        self.assertEqual(pypeg2.compose(result, output.Condition), condition)

    def test_complex_condition(self):
        condition = "(SIG(HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd) || (SIG(DNann1Lh55eZMEDXeYt59bzHbA3NJR46DeQYCS2qQdLV) && XHX(309BC5E644F797F53E5A2065EAF38A173437F2E6)))"
        result = pypeg2.parse(condition, output.Condition)
        self.assertEqual(result.left.left.pubkey, "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd")
        self.assertEqual(result.left.op.name, "||")
        self.assertEqual(result.left.right.left.pubkey, "DNann1Lh55eZMEDXeYt59bzHbA3NJR46DeQYCS2qQdLV")
        self.assertEqual(result.left.right.op.name, "&&")
        self.assertEqual(result.left.right.right.sha_hash, "309BC5E644F797F53E5A2065EAF38A173437F2E6")
        self.assertEqual(pypeg2.compose(result, output.Condition), condition)

    def test_csv_cltv_condition(self):
        condition = "(CSV(1654300) || (SIG(DNann1Lh55eZMEDXeYt59bzHbA3NJR46DeQYCS2qQdLV) && CLTV(2594024)))"
        result = pypeg2.parse(condition, output.Condition)
        self.assertEqual(result.left.left.time, "1654300")
        self.assertEqual(result.left.op.name, "||")
        self.assertEqual(result.left.right.left.pubkey, "DNann1Lh55eZMEDXeYt59bzHbA3NJR46DeQYCS2qQdLV")
        self.assertEqual(result.left.right.op.name, "&&")
        self.assertEqual(result.left.right.right.timestamp, "2594024")
        self.assertEqual(pypeg2.compose(result, output.Condition), condition)

    def test_instanciate_condition(self):
        inst = output.Condition.token(output.SIG.token("HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd"),
                                           output.Operator.token("||"),
                                           output.Condition.token(
                                               output.SIG.token("DNann1Lh55eZMEDXeYt59bzHbA3NJR46DeQYCS2qQdLV"),
                                               output.Operator.token("&&"),
                                               output.XHX.token("309BC5E644F797F53E5A2065EAF38A173437F2E6")
                                           ))
        condition = "(SIG(HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd) || (SIG(DNann1Lh55eZMEDXeYt59bzHbA3NJR46DeQYCS2qQdLV) && XHX(309BC5E644F797F53E5A2065EAF38A173437F2E6)))"
        inst = pypeg2.parse(condition, output.Condition)
        self.assertEqual(inst.left.left.pubkey, "HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd")
        self.assertEqual(inst.left.op.name, "||")
        self.assertEqual(inst.left.right.left.pubkey, "DNann1Lh55eZMEDXeYt59bzHbA3NJR46DeQYCS2qQdLV")
        self.assertEqual(inst.left.right.op.name, "&&")
        self.assertEqual(inst.left.right.right.sha_hash, "309BC5E644F797F53E5A2065EAF38A173437F2E6")
        self.assertEqual(pypeg2.compose(inst, output.Condition), condition)



#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Caner Candan <caner@candan.fr>, http://caner.candan.fr
#

from duniterpy.api.bma import API, logging, parse_response

logger = logging.getLogger("duniter/network")

URL_PATH = 'network'

PEERING_SCHEMA = {
        "type": "object",
        "properties": {
          "version": {
              "type": ["number", "string"]
          },
          "currency": {
              "type": "string"
          },
          "pubkey": {
              "type": "string"
          },
          "endpoints": {
              "type": "array",
              "items": {
                  "type": "string"
              }
          },
          "signature": {
              "type": "string"
          }
        },
        "required": ["version", "currency", "pubkey", "endpoints", "signature"]
    }

PEERS_SCHEMA = schema = {
        "type": ["object"],
        "properties": {
            "depth": {
                "type": "number"
            },
            "nodesCount": {
                "type": "number"
            },
            "leavesCount": {
                "type": "number"
            },
            "root": {
                "type": "string"
            },
            "hash": {
                "type": "string"
            },
            "value": {
                "type": "object",
                "properties": {
                    "version": {
                        "type": "string"
                    },
                    "currency": {
                        "type": "string"
                    },
                    "pubkey": {
                        "type": "string"
                    },
                    "endpoints": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "signature": {
                        "type": "string"
                    }
                },
                "required": ["version", "currency", "pubkey", "endpoints", "signature"]
            }
        },
        "oneOf": [
            {
                "required": ["depth", "nodesCount", "leavesCount", "root"]
            },
            {
                "required": ["hash", "value"]
            }
        ]
    }

async def peering(connection):
    """
    GET peering information about a peer

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    r = await client.requests_get('/peering')
    return await parse_response(r, PEERING_SCHEMA)

async def peers(connection, leaves=False, leaf=""):
    """
    GET peering entries of every node inside the currency network

    :param bool leaves: True if leaves should be requested
    :param str leaf: True if leaf should be requested
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    # GET Peers
    if leaves:
        r = await client.requests_get('/peering/peers', leaves=leaves)
    else:
        r = await client.requests_get('/peering/peers', leaf=leaf)

    return await parse_response(r, PEERS_SCHEMA)

async def peer(connection, entry=None, signature=None):
    """
    GET peering entries of every node inside the currency network

    :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
    :param duniterpy.documents.peer.Peer entry: Peer document
    :param str signature: Signature of the document issuer
    :rtype: dict
    """

    client = API(connection, URL_PATH)
    r = await client.requests_post('/peering/peers', entry=entry, signature=signature)
    return r

# async def status(connection):
#     """
# NOT DOCUMENTED IN BMA API DOCUMENTATION
#     POST a network status document to this node in order notify of its status
#
#     :param duniterpy.api.bma.ConnectionHandler connection: Connection handler instance
#     :param duniterpy.documents.peer.Peer entry: Peer document
#     :param str signature: Signature of the document issuer
#     :rtype: dict
#     """
#
#     async def __post__(self, session, **kwargs):
#         assert 'status' in kwargs
#         assert 'signature' in kwargs
#
#         r = await self.requests_post(session, '/status', **kwargs)
#         return r
